@@ -1,5 +1,5 @@
 ;;;; psgml-dtd.el --- DTD parser for SGML-editing mode with parsing support
-;; $Id: psgml-dtd.el,v 2.24 1998/11/14 12:02:15 lenst Exp $
+;; $Id: psgml-dtd.el,v 2.27 2001/04/02 22:04:44 lenst Exp $
 
 ;; Copyright (C) 1994 Lennart Staflin
 
@@ -168,7 +168,7 @@ Syntax: var dfa-expr &body forms"
     (setq l (sgml-state-opts s)
 	  nl (sgml-state-reqs s))
     (when yes
-      (sgml-debug "OPT redundant-1: sucess %s" (length yes))
+      (sgml-debug "OPT redundant-1: success %s" (length yes))
       (while (or l (setq l (prog1 nl (setq nl nil))))
 	(cond ((memq (sgml-move-dest (car l)) yes)
 	       (setf (sgml-move-dest (car l)) s)))
@@ -296,8 +296,7 @@ Syntax: var dfa-expr &body forms"
 	  "")
 	 (original-buffer		; Buffer (entity) where lit started
 	  (current-buffer))
-	 temp
-	 )
+	 temp)
     (cond
      ((or (sgml-parse-delim "LIT")
 	  (setq lita (sgml-parse-delim "LITA")))
@@ -316,14 +315,13 @@ Syntax: var dfa-expr &body forms"
 	      (t
 	       (setq value
 		     (concat value
-			     (buffer-substring
+			     (buffer-substring-no-properties
 			      (point)
 			      (progn (forward-char 1)
 				     (if lita
 					 (sgml-skip-upto ("LITA" "PERO" "CRO"))
 				       (sgml-skip-upto ("LIT" "PERO" "CRO")))
-				     (point)))))))
-	)
+				     (point))))))))
       value))))
 
 (defun sgml-check-parameter-literal ()
@@ -521,7 +519,7 @@ Case transformed for general names."
 	 (sgml-check-content-model))
 	(t
 	 ;; ANY, CDATA, RCDATA or EMPTY
-	 (let ((dc (intern (sgml-check-name)))) 
+	 (let ((dc (intern (sgml-check-case (sgml-check-name))))) 
 	   (cond ((eq dc 'ANY)
 		  (setq sgml-used-pcdata t))
 		 ((eq dc 'CDATA)
@@ -538,7 +536,7 @@ Case transformed for general names."
                                   "ANY, CDATA, RCDATA or EMPTY"))))
 	   dc))))
 
-(defun sgml-parse-exeption (type)
+(defun sgml-parse-exception (type)
   (sgml-skip-ps)
   (if (sgml-parse-char type)
       (if sgml-xml-p
@@ -562,8 +560,8 @@ Case transformed for general names."
 	 (etag-opt (sgml-parse-opt))
 	 (sgml-used-pcdata nil)
 	 (model (sgml-check-content))
-	 (exclusions (sgml-parse-exeption ?-))
-	 (inclusions (sgml-parse-exeption ?+)))
+	 (exclusions (sgml-parse-exception ?-))
+	 (inclusions (sgml-parse-exception ?+)))
     (sgml-before-eltype-modification)
     (while names
       (sgml-debug "Defining element %s" (car names))
@@ -618,7 +616,7 @@ Case transformed for general names."
 	    (let ((token (intern (sgml-check-case (sgml-check-name)))))
 	      (sgml-skip-ps)
 	      (when (and sgml-xml-p
-			 (memq token '(cdata sdata pi starttag endtag ms md)))
+			 (memq token '(CDATA SDATA PI STARTTAG ENDTAG MS MD)))
 		(sgml-error "XML forbids %s entities."
 			    (upcase (symbol-name token))))
 	      (cond
@@ -649,16 +647,14 @@ Case transformed for general names."
   (let ((type (sgml-parse-name))
 	(notation nil))
     (when type
-      (setq type (intern (downcase (sgml-check-case type))))
-      (when (and sgml-xml-p (memq type '(subdoc cdata sdata)))
+      (setq type (intern (sgml-check-case type)))
+      (when (and sgml-xml-p (memq type '(SUBDOC CDATA SDATA)))
 	(sgml-error "XML forbids %s entities."
 		    (upcase (symbol-name type))))
-      (cond ((eq type 'subdoc))
-	    ((memq type '(cdata ndata sdata))
+      (cond ((eq type 'SUBDOC))
+	    ((memq type '(CDATA NDATA SDATA))
 	     (sgml-skip-ps)
 	     (setq notation (sgml-parse-name))
-	     (when notation
-	       (setq notation (intern (downcase (sgml-check-case notation)))))
 	     ;;149.2+ data attribute specification
 	     ;;                      = 65 ps+, DSO,
 	     ;;                        31 attribute specification list,
@@ -669,8 +665,7 @@ Case transformed for general names."
 	       (sgml-parse-s)
 	       (sgml-check-delim DSC)))
 	    (t (sgml-error "Illegal entity type: %s" type))))
-    (cons type notation)
-    ))
+    (cons type notation)))
 
 
 ;;;; Parse doctype: Attlist
