@@ -1,5 +1,5 @@
 ;;;; psgml-other.el --- Part of SGML-editing mode with parsing support
-;; $Id: psgml-other.el,v 2.16 1999/06/16 18:24:15 lenst Exp $
+;; $Id: psgml-other.el,v 2.22 2001/11/04 23:49:02 lenst Exp $
 
 ;; Copyright (C) 1994 Lennart Staflin
 
@@ -30,6 +30,7 @@
 
 (require 'psgml)
 (require 'easymenu)
+(eval-when-compile (require 'cl))
 
 (defvar sgml-max-menu-size (/ (* (frame-height) 2) 3)
   "*Max number of entries in Tags and Entities menus before they are split
@@ -50,39 +51,47 @@ into several panes.")
   "Display a popup menu.
 ENTRIES is a list where every element has the form (STRING . VALUE) or
 STRING."
-  (x-popup-menu
-   event
-   (let ((menus (list (cons title entries))))
-     (cond
-      ((> (length entries) sgml-max-menu-size)
-       (setq menus
-	     (loop for i from 1 while entries
-		   collect
-		   (let ((submenu
-			  (subseq entries 0 (min (length entries)
-						 sgml-max-menu-size))))
-		     (setq entries (nthcdr sgml-max-menu-size entries))
-		     (cons
-		      (format "%s '%s'-'%s'"
-			      title
-			      (sgml-range-indicator (caar submenu))
-			      (sgml-range-indicator (caar (last submenu))))
-		      submenu))))))
-     (cons title menus))))
+  (let ((menus (sgml-split-long-menus (list (cons title entries)))))
+    (x-popup-menu event (cons title menus))))
+
 
 (defun sgml-range-indicator (string)
   (substring string
 	     0
 	     (min (length string) sgml-range-indicator-max-length)))
 
+
+(defun sgml-split-long-menus (menus)
+  (loop
+   for (title . entries) in menus
+   nconc
+   (cond
+    ((> (length entries) sgml-max-menu-size)
+     (loop for i from 1 while entries
+           collect
+           (let ((submenu (copy-sequence entries)))
+             (setcdr (nthcdr (1- (min (length entries) sgml-max-menu-size))
+                             submenu)
+                     nil)
+             (setq entries (nthcdr sgml-max-menu-size entries))
+             (cons
+              (format "%s '%s'.."
+                      title
+                      (sgml-range-indicator (caar submenu)))
+              submenu))))
+    (t
+     (list (cons title entries))))))
+
+
+
 (defun sgml-popup-multi-menu (event title menus)
   "Display a popup menu.
 MENUS is a list of menus on the form (TITLE ITEM1 ITEM2 ...).
 ITEM should have to form (STRING EXPR) or STRING.  The EXPR gets evaluated
 if the item is selected."
-  (nconc menus '(("---" "---")))	; Force x-popup-menu to use two level
-					; menu even if there is only one entry
-					; on the first level
+  (setq menus (sgml-split-long-menus menus))
+  (unless (cdr menus)
+    (setq menus (list (car menus) '("---" "---"))))
   (eval (car (x-popup-menu event (cons title menus)))))
 
 
@@ -111,11 +120,13 @@ if the item is selected."
     (cond
      (sgml-use-text-properties
       (let ((inhibit-read-only t)
-	    (after-change-function nil)	; obsolete variable
-	    (before-change-function nil) ; obsolete variable
-	    (after-change-functions nil)
-	    (before-change-functions nil))
-	(put-text-property start end 'face face)))
+            (after-change-functions nil)
+            (before-change-functions nil)
+            (buffer-undo-list t)
+            (deactivate-mark nil))
+	(put-text-property start end 'face face)
+        (when (< start end)
+          (put-text-property (1- end) end 'rear-nonsticky '(face)))))
      (t
       (let ((current (overlays-at start))
 	    (pos start)
@@ -126,7 +137,7 @@ if the item is selected."
 		      (eq type (overlay-get (car current) 'sgml-type)))
 		 (setq old-overlay (car current)))
 		((overlay-get (car current) 'sgml-type)
-		 (message "delov: %s" (overlay-get (car current) 'sgml-type))
+		 ;;(message "delov: %s" (overlay-get (car current) 'sgml-type))
 		 (delete-overlay (car current))))
 	  (setq current (cdr current)))
 	(while (< (setq pos (next-overlay-change pos))
