@@ -1,5 +1,5 @@
 ;;; psgml.el --- SGML-editing mode with parsing support
-;; $Id: psgml.el,v 2.48 1999/10/10 13:48:38 lenst Exp $
+;; $Id: psgml.el,v 2.59 2001/12/14 10:33:57 lenst Exp $
 
 ;; Copyright (C) 1993-1999 Lennart Staflin
 ;; Copyright (C) 1992 Free Software Foundation, Inc.
@@ -52,7 +52,7 @@
 
 ;;; Code:
 
-(defconst psgml-version "1.2.0"
+(defconst psgml-version "1.2.4"
   "Version of psgml package.")
 
 (defconst psgml-maintainer-address "lenst@lysator.liu.se")
@@ -72,7 +72,8 @@
   "Abbrev table in use in sgml-mode.")
 (define-abbrev-table 'sgml-mode-abbrev-table ())
 
-(defvar sgml-running-lucid (string-match "Lucid" emacs-version))
+(eval-and-compile
+  (defvar sgml-running-lucid (string-match "Lucid" emacs-version)))
 
 (defvar sgml-xml-p nil
   "Is this an XML document?")
@@ -80,6 +81,10 @@
 
 ;;; User settable options:
 
+(defvar sgml-insert-defaulted-attributes nil
+  "*Controls whether defaulted attributes (not #FIXED) are inserted explicitly
+or not. nil means don't insert, t means insert.")
+
 (defvar sgml-insert-missing-element-comment t
   "*If true, and sgml-auto-insert-required-elements also true,
 `sgml-insert-element' will insert a comment if there is an element required
@@ -100,8 +105,8 @@ Setting this variable automatically makes it local to the current buffer.")
 (defvar sgml-system-identifiers-are-preferred nil
   "*If nil, PSGML will look up external entities by searching the catalogs
 in `sgml-local-catalogs' and `sgml-catalog-files' and only if the
-entity is not found in the catalogs will a given system identifer be
-used. If the variable is non-nil and a system identifer is given, the
+entity is not found in the catalogs will a given system identifier be
+used. If the variable is non-nil and a system identifier is given, the
 system identifier will be used for the entity. If no system identifier
 is given the catalogs will searched.")
 
@@ -170,7 +175,7 @@ hierarchy. The variable should have the form
 
   (PARENT-FILE CONTEXT-ELEMENT* TOP-ELEMENT (HAS-SEEN-ELEMENT*)?)
 
-PARENT-FILE	is a string, the name of the file contatining the
+PARENT-FILE	is a string, the name of the file containing the
 		document entity.
 CONTEXT-ELEMENT is a string, that is the name of an element type.
 		It can occur 0 or more times and is used to set up
@@ -285,7 +290,7 @@ where and has content model ANY.")
   "*If non-nil, out of date compiled DTDs will be automatically recompiled.
 If the value is `ask', PSGML will ask before recompiling. A `nil'
 value will cause PSGML to silently load an out of date compiled DTD.
-A DTD that referes to undefined external entities is always out of
+A DTD that refers to undefined external entities is always out of
 date, thus in such case it can be useful to set this variable to
 `nil'.")
 (put 'sgml-recompile-out-of-date-cdtd 'sgml-type '(("No" . nil)
@@ -353,7 +358,7 @@ identifier.")
 
 (defvar sgml-local-catalogs nil
 "*A list of SGML entity catalogs to be searched first when parsing the buffer.
-This is used in addtion to `sgml-catalog-files',  and `sgml-public-map'.
+This is used in addition to `sgml-catalog-files',  and `sgml-public-map'.
 This variable is automatically local to the buffer.")
 (make-variable-buffer-local 'sgml-local-catalogs)
 (put 'sgml-local-catalogs 'sgml-type 'file-list)
@@ -404,7 +409,7 @@ use `setq-default' to set it to a value that is shared among buffers.")
 
 (defvar sgml-custom-markup nil
   "*Menu entries to be added to the Markup menu.
-The value should be a list of lists of two strings.  The first is a
+The value should be a list of lists of two strings.  The first
 string is the menu line and the second string is the text inserted
 when the menu item is chosen.  The second string can contain a \\r
 where the cursor should be left.  Also if a selection is made
@@ -420,12 +425,12 @@ Example:
 
 (defvar sgml-custom-dtd nil
   "Menu entries to be added to the DTD menu.
-The value should be a list of entrys to be added to the DTD menu.
+The value should be a list of entries to be added to the DTD menu.
 Every entry should be a list. The first element of the entry is a string
 used as the menu entry.  The second element is a string containing a
 doctype declaration (this can be nil if no doctype).  The rest of the
 list should be a list of variables and values.  For backward
-compatibility a singel string instead of a variable is assigned to
+compatibility a single string instead of a variable is assigned to
 sgml-default-dtd-file.  All variables are made buffer local and are also
 added to the buffers local variables list.
 
@@ -481,7 +486,7 @@ These file names will serve as the arguments to the `sgml-validate-command'
 format control string instead of the defaults.")
 
 (defvar sgml-validate-error-regexps
-  '((".*:\\(.+\\):\\([0-9]+\\):\\([0-9]+\\):[EX]: " 1 2 3)
+  '((".*:\\(.+\\):\\([0-9]+\\):\\([0-9]+\\):[EXW]: " 1 2 3)
     ("\\(error\\|warning\\) at \\([^,]+\\), line \\([0-9]+\\)" 2 3)
     ("\n[a-zA-Z]?:?[^0-9 \n\t:]+:[ \t]*\\([^ \n\t:]+\\):\
 \\([0-9]+\\):\\(\\([0-9]+\\)[: \t]\\)?" 1 2 4))
@@ -761,6 +766,14 @@ as that may change."
 	 'sgml-tag-region-if-active
          'sgml-use-text-properties
 	 ))))
+
+;;;; SGML mode: syntax table
+
+(defvar sgml-mode-syntax-table
+  (let ((s (copy-syntax-table text-mode-syntax-table)))
+    (modify-syntax-entry ?< "."  s)
+    (modify-syntax-entry ?> "."  s)
+    s))
 
 
 ;;;; SGML mode: keys and menus
@@ -818,7 +831,7 @@ as that may change."
 (define-key sgml-mode-map "\e\C-k"   'sgml-kill-element)
 (define-key sgml-mode-map "\e\C-@"   'sgml-mark-element)
 ;;(define-key sgml-mode-map [?\M-\C-\ ] 'sgml-mark-element)
-(define-key sgml-mode-map "\e\C-h"   'sgml-mark-current-element)
+(define-key sgml-mode-map [(meta control h)] 'sgml-mark-current-element)
 (define-key sgml-mode-map "\e\C-t"   'sgml-transpose-element)
 (define-key sgml-mode-map "\M-\t"    'sgml-complete)
 
@@ -972,12 +985,15 @@ as that may change."
     (when (or (null last-values)
               (sgml-any-option-changed (cdr last-values)
                                        option-vars))
-      (easy-menu-change '("SGML") menuname
+      (condition-case err
+          (easy-menu-change '("SGML") menuname
                       (nconc (sgml-options-menu-items option-vars)
                              (if save-func
                                  (list "---"
                                        (vector (format "Save %s" menuname)
                                                save-func t)))))
+        (error
+         (message "Error in update menu: %s" err)))
       (unless last-values
         (setq last-values (cons menuname nil))
         (push last-values sgml-last-options-menu-values))
@@ -1078,7 +1094,7 @@ To tag a region with the mouse, use transient mark mode or secondary selection.
 
 Structure editing:
 \\[sgml-backward-element]  Moves backwards over the previous element.
-\\[sgml-forward-element]  Moves forward over the nex element.
+\\[sgml-forward-element]  Moves forward over the next element.
 \\[sgml-down-element]  Move forward and down one level in the element structure.
 \\[sgml-backward-up-element]  Move backward out of this element level.
 \\[sgml-beginning-of-element]  Move to after the start tag of the current element.
@@ -1100,7 +1116,7 @@ Folding and unfolding
 User options:
 
 sgml-omittag  Set this to reflect OMITTAG in the SGML declaration.
-sgml-shortag  Set this to reflect SHORTTAG in the SGML declaration.
+sgml-shorttag  Set this to reflect SHORTTAG in the SGML declaration.
 sgml-namecase-general  Set this to reflect NAMECASE GENERAL in the SGML declaration.
 sgml-auto-insert-required-elements  If non-nil, automatically insert required 
 	elements in the content of an inserted element.
@@ -1119,7 +1135,7 @@ sgml-minimize-attributes  Determines minimization of attributes inserted by
 	edit-attributes.
 sgml-normalize-trims  If non-nil, sgml-normalize will trim off white space 
 	from end of element when adding end tag.
-sgml-indent-step  How much to increament indent for every element level.
+sgml-indent-step  How much to increment indent for every element level.
 sgml-indent-data  If non-nil, indent in data/mixed context also.
 sgml-set-face     If non-nil, psgml will set the face of parsed markup.
 sgml-markup-faces The faces used when the above variable is non-nil.
@@ -1149,7 +1165,7 @@ All bindings:
   (set (make-local-variable 'paragraph-start)
        paragraph-separate)
 
-  (set-syntax-table text-mode-syntax-table)
+  (set-syntax-table sgml-mode-syntax-table)
   (make-local-variable 'comment-start)
   (setq comment-start "<!-- ")
   (make-local-variable 'comment-end)
@@ -1209,6 +1225,7 @@ All bindings:
   (setq sgml-minimize-attributes nil)
   (setq sgml-always-quote-attributes t)
   (setq sgml-validate-command "nsgmls -wxml -s %s %s")
+  ;; FIXME: why conditional?
   (unless sgml-declaration
     (make-local-variable 'sgml-declaration)
     (setq sgml-declaration sgml-xml-declaration)))
@@ -1421,6 +1438,7 @@ and move to the line in the SGML document that caused it."
 	  nil nil)
 (autoload 'sgml-indent-line "psgml-edit" nil)
 (autoload 'sgml-element-endable-p "psgml-edit" nil)
+(autoload 'sgml-do-set-option "psgml-edit" nil)
 
 ;;; Generated by sgml-build-autoloads
 
@@ -1437,7 +1455,7 @@ That is move to before the start-tag or where a start-tag is implied." t)
 That is move to after the end-tag or where an end-tag is implied." t)
 (autoload 'sgml-forward-element "psgml-edit" "Move forward over next element." t)
 (autoload 'sgml-backward-element "psgml-edit" "Move backward over previous element at this level.
-With implied tags this is ambigous." t)
+With implied tags this is ambiguous." t)
 (autoload 'sgml-down-element "psgml-edit" "Move forward and down one level in the element structure." t)
 (autoload 'sgml-kill-element "psgml-edit" "Kill the element following the cursor." t)
 (autoload 'sgml-transpose-element "psgml-edit" "Interchange element before point with element after point, leave point after." t)
@@ -1467,7 +1485,7 @@ This uses the selective display feature." t)
 (autoload 'sgml-insert-tag "psgml-edit" "Insert a tag, reading tag name in minibuffer with completion.
 If the variable sgml-balanced-tag-edit is t, also inserts the
 corresponding end tag. If sgml-leave-point-after-insert is t, the point
-is left after the inserted tag(s), unless the element has som required
+is left after the inserted tag(s), unless the element has some required
 content.  If sgml-leave-point-after-insert is nil the point is left
 after the first tag inserted." t)
 (autoload 'sgml-insert-element "psgml-edit" "Reads element name from minibuffer and inserts start and end tags." t)
@@ -1479,10 +1497,10 @@ If repeated, the containing element will be split before the beginning
 of then current element." t)
 (autoload 'sgml-custom-dtd "psgml-edit" "Insert a DTD declaration from the sgml-custom-dtd alist." t)
 (autoload 'sgml-custom-markup "psgml-edit" "Insert markup from the sgml-custom-markup alist." t)
-(autoload 'sgml-tags-menu "psgml-edit" "Pop up a menu with valid tags and insert the choosen tag.
+(autoload 'sgml-tags-menu "psgml-edit" "Pop up a menu with valid tags and insert the chosen tag.
 If the variable sgml-balanced-tag-edit is t, also inserts the
 corresponding end tag. If sgml-leave-point-after-insert is t, the point
-is left after the inserted tag(s), unless the element has som required
+is left after the inserted tag(s), unless the element has some required
 content.  If sgml-leave-point-after-insert is nil the point is left
 after the first tag inserted." t)
 (autoload 'sgml-element-menu "psgml-edit" "Pop up a menu with valid elements and insert choice.
@@ -1493,8 +1511,8 @@ tag inserted." t)
 (autoload 'sgml-tag-region-menu "psgml-edit" "Pop up a menu with valid elements and tag current region with the choice." t)
 (autoload 'sgml-entities-menu "psgml-edit" nil t)
 (autoload 'sgml-attrib-menu "psgml-edit" "Pop up a menu of the attributes of the current element
-\(or the element whith start-tag before point)." t)
-(autoload 'sgml-fill-element "psgml-edit" "Fill bigest enclosing element with mixed content.
+\(or the element with start-tag before point)." t)
+(autoload 'sgml-fill-element "psgml-edit" "Fill biggest enclosing element with mixed content.
 If current element has pure element content, recursively fill the
 subelements." t)
 (autoload 'sgml-edit-attributes "psgml-edit" "Edit attributes of current element.
@@ -1511,14 +1529,14 @@ Editing is done in a separate window." t)
 (autoload 'sgml-show-attributes "psgml-edit" "Show all attribute specifications in the buffer." t)
 (autoload 'sgml-expand-all-shortrefs "psgml-edit" "Expand all short references in the buffer.
 Short references to text entities are expanded to the replacement text
-of the entity other short references are expanded into general entity
-references.  If argument, TO-ENTITY, is non-nil, or if called
-interactive with numeric prefix argument, all short references are
-replaced by generaly entity references." t)
+of the entity; other short references are expanded into general entity
+references.  If argument TO-ENTITY is non-nil, or if called
+interactively with a numeric prefix argument, all short references are
+replaced by general entity references." t)
 (autoload 'sgml-normalize "psgml-edit" "Normalize buffer by filling in omitted tags and expanding empty tags.
 Argument TO-ENTITY controls how short references are expanded as with
 `sgml-expand-all-shortrefs'.  An optional argument ELEMENT can be the
-element to normalize insted of the whole buffer, if used no short
+element to normalize instead of the whole buffer, if used no short
 references will be expanded." t)
 (autoload 'sgml-normalize-element "psgml-edit" nil t)
 (autoload 'sgml-make-character-reference "psgml-edit" "Convert character after point into a character reference.
@@ -1555,10 +1573,11 @@ otherwise it will be added at the first legal position." t)
 (provide 'psgml)
 (provide 'sgml-mode)
 
-(cond
- (sgml-running-lucid
-  (require 'psgml-lucid))
- (t
-  (require 'psgml-other)))
+(eval-and-compile
+  (cond
+   (sgml-running-lucid
+    (require 'psgml-lucid))
+   (t
+    (require 'psgml-other))))
 
 ;;; psgml.el ends here
